/* ======================================================================

   Swirly - a psychedelic screen saver for the SGI

   Copyright 1998 Nathaniel Bletter.  All rights reserved.
   RESTRICTED RIGHTS NOTICE.  This is an unpublished work of
   and is not to be used, copied or disclosed without written permission
   of the author.

   ======================================================================
*/

#include <gl/gl.h>
#include <gl/device.h>
#include <sys/times.h>
#include <sys/types.h>
#include <math.h>
#include <time.h>
#include <stdio.h>

#define RAMPBASE 64
#define RAMPSIZE 128
#define RAMPSTEP (255 / (RAMPSIZE - 1))
#define RAMPCOLOR(fract)   ((Colorindex) ((fract) * (RAMPSIZE -1 ) + 0.5) + RAMPBASE)

#define MAXNODES 100
#define RNDBOX 10
#define RNDSIZE (RNDBOX*RNDBOX)

int quit = FALSE;

typedef struct nodetype {
  int pos[2];
  float aperiod;
  float spiral;
  float wave;
  int dir;
} nodetype;

int ncols;
int atime;
int rtime;
int numnodes;
nodetype node[MAXNODES];
long winx, winy;
int rndpos[RNDSIZE][2];

float sc[4][3] = {
  {0.0, 0.0, 0.0},
  {1.0, 0.0, 0.0},
  {1.0, 1.0, 0.0},
  {0.0, 1.0, 0.0}
};

void makemap();
void makenodes(int);
void drawnodes();
void animate();
void makerndpos();

main(int argc, char **argv)
{
  short val;
  Device dev;
  int i;
  time_t ti;
  int cnodes;
  short saver = FALSE;
  short value;
  int que;

  cnodes = 4;
  ncols = 5;
  atime = 10;
  rtime = 0;
  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-s") == 0) 
      saver = TRUE;
    else if (strcmp(argv[i], "-c") == 0 && argc > i+1) {
      ncols = atoi(argv[i+1]);
      i++;
    }
    else if (strcmp(argv[i], "-a") == 0 && argc > i+1) {
      atime = atoi(argv[i+1]);
      i++;
    }
    else if (strcmp(argv[i], "-r") == 0 && argc > i+1) {
      rtime = atoi(argv[i+1]);
      i++;
    }
    else if (argv[i][0] == '-') {
      fprintf(stderr, "Usage: %s [-s] [-c n] [-a n] [-r n] [#]\n", argv[0]);
      fprintf(stderr, "   -s      Run in screen saver mode (fullscreen, no window border)\n");
      fprintf(stderr, "   -c	n		Make a color between n base colors (default 5)\n");
      fprintf(stderr, "   -a	t		animation speed (0 = fastest, 10 = normal)\n");
      fprintf(stderr, "   -r	t		processing speed, slower means more time for other processes (0 = fastest <default>)\n");
      fprintf(stderr, "   #       Number of nodes to draw. Default is 4.\n");
      exit(-1);
    }
    else
      cnodes = atoi(argv[i]);
  }

  winx = getgdesc(GD_XPMAX);
  winy = getgdesc(GD_YPMAX);
  if (saver) {
    prefsize(winx, winy);
    noborder();
  }
  foreground();
  winopen("ramp");
  getsize(&winx, &winy);
  cmode();
  gconfig();

  qdevice(DEPTHCHANGE);
  qdevice(WINTHAW);
  qdevice(REDRAW);
/*  qdevice(WINQUIT);*/
  qdevice(ESCKEY);

  ortho2(-0.5, winx-0.5, -0.5, winy-0.5);

  /* get random seed from time */
  ti = time(NULL);
  srand48((long) (ti << 2));

  makerndpos();       /* create a random ordering of the screen points, so */
                      /* swirly pattern emerges evenly across the sreen */
  makenodes(cnodes);  /* create the field nodes with each of their random parameters */

  makemap(); /* create color map */
  drawnodes(); /* draw the nodes once */

  while (!quit) {
    if (que = qtest()) {
      dev = qread(&value); 
      if (dev == ESCKEY || dev == WINQUIT)
	quit = TRUE;
    }
    if ((time(NULL) - ti) > 30 || (que && (dev == REDRAW || dev == WINTHAW))) {
      if (qtest() && (dev = REDRAW || dev == WINTHAW)) {
	/* window got popped */
	singlebuffer();
	gconfig();
      }
      getsize(&winx, &winy);
      ortho2(-0.5, winx-0.5, -0.5, winy-0.5);
      viewport(0, winx-1, 0, winy-1);
      makenodes(numnodes);
      drawnodes();
      ti = time(NULL);
    }
    animate();
  }

/*  sleep(10);*/
  gexit();
  return 0;
}

void makenodes(int num)
{
  int i;

  if (num < MAXNODES)
    numnodes = num;
  else
    numnodes = MAXNODES - 1;
      
  for (i = 0; i < num; i++) {
    /* screen position of node */
    node[i].pos[0] = winx*drand48();
    node[i].pos[1] = winy*drand48();
    /* angular period  of arms */
    node[i].aperiod =  RAMPSIZE * floor(drand48()*4.5 + 1) / (2*M_PI);
    /* strength of ripples */
    node[i].wave = M_PI / 2.0 * drand48();
    /* effect of spiral */
    node[i].spiral = 4*M_PI / log(winx/4.0) * drand48();
    /* direction of rotation */
    node[i].dir = i%2 ? 1 : -1;
  }
}


/* make a pseudo random array of points, all visited once */
/* this pattern covers the screen evenly */
void makerndpos()
{
  int i, j, k, a;
  int tmp;

  /* make it ordered */
  for (i = 0; i < RNDBOX; i++) {
    for (j = 0; j < RNDBOX; j++) {
      rndpos[i*RNDBOX+j][0] = i;
      rndpos[i*RNDBOX+j][1] = j;
    }
  }
  
  /* Then shuffle it */
  for (i = 0; i < RNDSIZE*10; i++) {
    j = drand48() * RNDSIZE;
    k = drand48() * RNDSIZE;
    for (a = 0; a < 2; a++) {
      tmp = rndpos[j][a];
      rndpos[j][a] = rndpos[k][a];
      rndpos[k][a] = tmp;
    }
  }
}


/* create a color map that smoothly interpolates between ncols colors */
void makemap()
{
#define MAXCOLS 20
#define PI4 12.56637
  float cols[MAXCOLS][3];
  int i, c;
  float step;
  float m[3];
  int ccol, sec, nextcol;

  /* make colors */
  for(i = 0; i<ncols; i++) {
    for (c = 0; c < 3; c++) 
      cols[i][c] = 255 * drand48();
  }

  /* make ramp */
  step = (float) RAMPSIZE / (float)ncols;
  i = RAMPBASE;
  for (ccol = 0; ccol <= ncols; ccol++) {
    nextcol = (ccol + 1) % ncols;
    for (c = 0; c < 3; c++)
      m[c] = (cols[nextcol][c] - cols[ccol][c]) / step;

    for(sec = 0; sec < step; sec++) {
      mapcolor(i, m[0] * sec + cols[ccol][0], m[1] * sec + cols[ccol][1], m[2] * sec + cols[ccol][2]);
      i++;
    }
  }

}

void drawnodes()
{
  short val;
  Device dev;
  int i, x, y;
  float f, rad, r2, th;
  float dx, dy;
  short p[2];
  float coef, eff, thcoef, coscoef, decay;
  float v[4][2];
  int sizex, sizey;
  int ix, iy, s, sx, sy;
  int tx, ty;
  int ind;
#define PERIOD 2.0
#define APERIOD 3.0
#define WIGPERIOD 75.0

  coef = 2*M_PI / log(winx/4.0);
  thcoef =  RAMPSIZE * APERIOD / (2*M_PI);
  coscoef =  M_PI * WIGPERIOD/(float)winx;
  decay = 1.0/(winx/10.0);

 start:
  /* draw simple background polygon before we start */
  for (i = 0; i < 4; i++) {
    v[i][0] = winx * sc[i][0];
    v[i][1] = winy * sc[i][1];
  }

  color(BLACK);
  clear();
  bgnpolygon();
  colorf(RAMPCOLOR(0.0));
  v2f(v[0]);
  colorf(RAMPCOLOR(0.25));
  v2f(v[1]);
  colorf(RAMPCOLOR(1.0));
  v2f(v[2]);
  colorf(RAMPCOLOR(0.5));
  v2f(v[3]);
  endpolygon();

  bgnpoint();  

  /* how many randomized tiles across the screen? */
  sizex = ceil(winx / (float)RNDBOX);
  sizey = ceil(winy / (float)RNDBOX);

  for (s = 0; s < RNDSIZE; s++) {
    for (ix = 0, sx = 0; ix < sizex; ix++, sx += RNDBOX) {
      for (iy = 0, sy = 0; iy < sizey; iy++, sy += RNDBOX) {
	ind = (s+ix*iy*(long)8429) % RNDSIZE;
	x = sx + rndpos[ind][0];
	y = sy + rndpos[ind][1];

	if (x > winx || y > winy)
	  continue;

	f = 0; /* field value */

	for (i = 0; i < numnodes; i++) {
	  dx = x - node[i].pos[0];
	  dy = y - node[i].pos[1];
	  rad = sqrtf(dx*dx + dy*dy);
	  eff = 1/powf(decay*rad + 1, 4);
	  th = node[i].dir* (node[i].spiral * flog(rad+0.01) + (atan2f(dy, dx))+ eff * node[i].wave * sinf(rad * coscoef));

	  f += (th + PI4) * node[i].aperiod;
	}
	f += drand48()*5.0 - 2.5; /* add some graininess for texture */
	/* finally draw point */
	color(fmodf(f, RAMPSIZE) + RAMPBASE);
	p[1] = y;
	p[0] = x;
	v2s(p);
	/* do color map animation */
	animate();
	sginap(rtime); /* rest so we don't take up the entire processor */
      }
    }
    if (qtest())
      if ((dev = qread(&val)) == REDRAW || dev == WINTHAW) {
	/* window popped, start over */
	endpoint();
	getsize(&winx, &winy);
	ortho2(-0.5, winx-0.5, -0.5, winy-0.5);
	viewport(0, winx-1, 0, winy-1);
	makenodes(numnodes);
	singlebuffer();
	gconfig();
	goto start;
      }
      else if (dev == ESCKEY || dev == WINQUIT) {
	quit = TRUE;
	return;
      }
  }

  endpoint();
  makemap();
  doublebuffer();
  swapbuffers();
  gconfig();
  
}


/* cycle color map */
void animate()
{
  static clock_t last = -10000;
  clock_t ctime;
  struct tms buf;
  int i;
  short r, g, b, or, og, ob;

  /* do animation at constant time intervals */
  if ((ctime = times(&buf)) - last < atime)
    return;
  else 
    last = ctime;

  getmcolor(RAMPSIZE - 1 + RAMPBASE, &or, &og, &ob);
  for (i = 0; i<RAMPSIZE; i++) {
    getmcolor(i + RAMPBASE, &r, &g, &b);
    mapcolor(i+RAMPBASE, or, og, ob);
    or = r;
    og = g;
    ob =b;
  }
  gflush();
}
